import * as plugins from './smartproxy.plugins.js';

export interface IDomainConfig {
  domain: string;         // glob pattern for domain
  allowedIPs: string[];   // glob patterns for IPs allowed to access this domain
  targetIP?: string;      // Optional target IP for this domain
}

export interface IProxySettings extends plugins.tls.TlsOptions {
  // Port configuration
  fromPort: number;
  toPort: number;
  toHost?: string;        // Target host to proxy to, defaults to 'localhost'

  // Domain and security settings
  domains: IDomainConfig[];
  sniEnabled?: boolean;
  defaultAllowedIPs?: string[]; // Optional default IP patterns if no matching domain found
  preserveSourceIP?: boolean;   // Whether to preserve the client's source IP when proxying
}

/**
 * Extract SNI (Server Name Indication) from a TLS ClientHello packet.
 * Returns the server name if found, or undefined.
 */
function extractSNI(buffer: Buffer): string | undefined {
  let offset = 0;
  // We need at least 5 bytes for the record header.
  if (buffer.length < 5) {
    return undefined;
  }

  // TLS record header
  const recordType = buffer.readUInt8(0);
  if (recordType !== 22) { // 22 = handshake
    return undefined;
  }
  // Read record length
  const recordLength = buffer.readUInt16BE(3);
  if (buffer.length < 5 + recordLength) {
    // Not all data arrived yet; in production you might need to accumulate more data.
    return undefined;
  }

  offset = 5;
  // Handshake message type should be 1 for ClientHello.
  const handshakeType = buffer.readUInt8(offset);
  if (handshakeType !== 1) {
    return undefined;
  }
  // Skip handshake header (1 byte type + 3 bytes length)
  offset += 4;

  // Skip client version (2 bytes) and random (32 bytes)
  offset += 2 + 32;

  // Session ID
  const sessionIDLength = buffer.readUInt8(offset);
  offset += 1 + sessionIDLength;

  // Cipher suites
  const cipherSuitesLength = buffer.readUInt16BE(offset);
  offset += 2 + cipherSuitesLength;

  // Compression methods
  const compressionMethodsLength = buffer.readUInt8(offset);
  offset += 1 + compressionMethodsLength;

  // Extensions length
  if (offset + 2 > buffer.length) {
    return undefined;
  }
  const extensionsLength = buffer.readUInt16BE(offset);
  offset += 2;
  const extensionsEnd = offset + extensionsLength;

  // Iterate over extensions
  while (offset + 4 <= extensionsEnd) {
    const extensionType = buffer.readUInt16BE(offset);
    const extensionLength = buffer.readUInt16BE(offset + 2);
    offset += 4;

    // Check for SNI extension (type 0)
    if (extensionType === 0x0000) {
      // SNI extension: first 2 bytes are the SNI list length.
      if (offset + 2 > buffer.length) {
        return undefined;
      }
      const sniListLength = buffer.readUInt16BE(offset);
      offset += 2;
      const sniListEnd = offset + sniListLength;
      // Loop through the list; typically there is one entry.
      while (offset + 3 < sniListEnd) {
        const nameType = buffer.readUInt8(offset);
        offset++;
        const nameLen = buffer.readUInt16BE(offset);
        offset += 2;
        if (nameType === 0) { // host_name
          if (offset + nameLen > buffer.length) {
            return undefined;
          }
          const serverName = buffer.toString('utf8', offset, offset + nameLen);
          return serverName;
        }
        offset += nameLen;
      }
      break;
    } else {
      offset += extensionLength;
    }
  }
  return undefined;
}

export class PortProxy {
  netServer: plugins.net.Server;
  settings: IProxySettings;

  constructor(settings: IProxySettings) {
    this.settings = {
      ...settings,
      toHost: settings.toHost || 'localhost'
    };
  }

  public async start() {
    const cleanUpSockets = (from: plugins.net.Socket, to: plugins.net.Socket) => {
      from.end();
      to.end();
      from.removeAllListeners();
      to.removeAllListeners();
      from.unpipe();
      to.unpipe();
      from.destroy();
      to.destroy();
    };

    const normalizeIP = (ip: string): string[] => {
      // Handle IPv4-mapped IPv6 addresses
      if (ip.startsWith('::ffff:')) {
        const ipv4 = ip.slice(7); // Remove '::ffff:' prefix
        return [ip, ipv4];
      }
      // Handle IPv4 addresses by adding IPv4-mapped IPv6 variant
      if (/^\d{1,3}(\.\d{1,3}){3}$/.test(ip)) {
        return [ip, `::ffff:${ip}`];
      }
      return [ip];
    };

    const isAllowed = (value: string, patterns: string[]): boolean => {
      // Expand patterns to include both IPv4 and IPv6 variants
      const expandedPatterns = patterns.flatMap(normalizeIP);
      // Check if any variant of the IP matches any expanded pattern
      return normalizeIP(value).some(ip =>
        expandedPatterns.some(pattern => plugins.minimatch(ip, pattern))
      );
    };

    const findMatchingDomain = (serverName: string): IDomainConfig | undefined => {
      return this.settings.domains.find(config => plugins.minimatch(serverName, config.domain));
    };

    // Always create a plain net server for TLS passthrough.
    this.netServer = plugins.net.createServer((socket: plugins.net.Socket) => {
      const remoteIP = socket.remoteAddress || '';
      
      // If SNI is enabled, we peek at the first chunk to extract the SNI.
      if (this.settings.sniEnabled) {
        socket.once('data', (chunk: Buffer) => {
          // Try to extract the server name from the ClientHello.
          const serverName = extractSNI(chunk) || '';
          console.log(`Received connection from ${remoteIP} with SNI: ${serverName}`);

          // Check if the IP is allowed by default.
          const isDefaultAllowed = this.settings.defaultAllowedIPs && isAllowed(remoteIP, this.settings.defaultAllowedIPs);
          if (!isDefaultAllowed && serverName) {
            const domainConfig = findMatchingDomain(serverName);
            if (!domainConfig) {
              console.log(`Connection rejected: No matching domain config for ${serverName} from IP ${remoteIP}`);
              socket.end();
              return;
            }
            if (!isAllowed(remoteIP, domainConfig.allowedIPs)) {
              console.log(`Connection rejected: IP ${remoteIP} not allowed for domain ${serverName}`);
              socket.end();
              return;
            }
          } else if (!isDefaultAllowed && !serverName) {
            console.log(`Connection rejected: No SNI and IP ${remoteIP} not in default allowed list`);
            socket.end();
            return;
          } else {
            console.log(`Connection allowed: IP ${remoteIP} is in default allowed list`);
          }

          // Determine target host.
          const domainConfig = serverName ? findMatchingDomain(serverName) : undefined;
          const targetHost = domainConfig?.targetIP || this.settings.toHost!;

          // Create connection options.
          const connectionOptions: plugins.net.NetConnectOpts = {
            host: targetHost,
            port: this.settings.toPort,
          };
          if (this.settings.preserveSourceIP) {
            connectionOptions.localAddress = remoteIP.replace('::ffff:', '');
          }

          const to = plugins.net.connect(connectionOptions);
          console.log(`Connection established: ${remoteIP} -> ${targetHost}:${this.settings.toPort}${serverName ? ` (SNI: ${serverName})` : ''}`);

          // Unshift the data chunk back so that the TLS handshake can complete at the backend.
          socket.unshift(chunk);
          socket.setTimeout(120000);
          socket.pipe(to);
          to.pipe(socket);

          const errorHandler = () => {
            cleanUpSockets(socket, to);
          };
          socket.on('error', errorHandler);
          to.on('error', errorHandler);
          socket.on('close', errorHandler);
          to.on('close', errorHandler);
          socket.on('timeout', errorHandler);
          to.on('timeout', errorHandler);
          socket.on('end', errorHandler);
          to.on('end', errorHandler);
        });
      } else {
        // If SNI is not enabled, use defaultAllowedIPs check.
        if (!this.settings.defaultAllowedIPs || !isAllowed(remoteIP, this.settings.defaultAllowedIPs)) {
          console.log(`Connection rejected: IP ${remoteIP} not allowed for non-SNI connection`);
          socket.end();
          return;
        }
        const targetHost = this.settings.toHost!;
        const connectionOptions: plugins.net.NetConnectOpts = {
          host: targetHost,
          port: this.settings.toPort,
        };
        if (this.settings.preserveSourceIP) {
          connectionOptions.localAddress = remoteIP.replace('::ffff:', '');
        }
        const to = plugins.net.connect(connectionOptions);
        console.log(`Connection established: ${remoteIP} -> ${targetHost}:${this.settings.toPort}`);
        socket.setTimeout(120000);
        socket.pipe(to);
        to.pipe(socket);
        const errorHandler = () => {
          cleanUpSockets(socket, to);
        };
        socket.on('error', errorHandler);
        to.on('error', errorHandler);
        socket.on('close', errorHandler);
        to.on('close', errorHandler);
        socket.on('timeout', errorHandler);
        to.on('timeout', errorHandler);
        socket.on('end', errorHandler);
        to.on('end', errorHandler);
      }
    })
    .on('error', (err: Error) => {
      console.log(`Server Error: ${err.message}`);
    })
    .listen(this.settings.fromPort, () => {
      console.log(`PortProxy -> OK: Now listening on port ${this.settings.fromPort}${this.settings.sniEnabled ? ' (SNI passthrough enabled)' : ''}`);
    });
  }

  public async stop() {
    const done = plugins.smartpromise.defer();
    this.netServer.close(() => {
      done.resolve();
    });
    await done.promise;
  }
}